open Base
open Exceptions

module type Intf = sig
  (** [with_db ?read_only filename fn] opens a database from [filename], applies
      a callback [fn] to it and closes it after [fn] returns *)
  val with_db: ?read_only:bool -> string -> (Sqlite3.db -> 'a) -> 'a

  (** [exec_only ?params db query] creates a new Sqlite statement from [query],
      binds [?params] to it and executes it without returning any result *)
  val exec_only:
    ?params:(string * Sqlite3.Data.t) list
    -> Sqlite3.db
    -> string
    -> unit

  (** [exec ?params db query] creates a new Sqlite statement from [query],
      binds [?params] to it and executes it reading every resulting row and
      converting it using [of_sql_row] function from the given module. *)
  val exec_fetch:
    (module From_sql.Sig with type t = 't)
    -> ?params:(string * Sqlite3.Data.t) list
    -> Sqlite3.db
    -> string
    -> 't array
end

module Impl: Intf = struct
  let with_db ?(read_only=false) filename fn =
    try
      let db = match read_only with
        | true -> Sqlite3.db_open ~mode:`READONLY filename
        | false -> Sqlite3.db_open filename in
      Exn.protect
        ~f:begin fun () ->
          try
            fn db
          with
          | Sqlite3.Error reason ->
            Sqlite_error (Sqlite3.Rc.ERROR, reason) |> raise
        end
        ~finally:(fun () -> ignore (Sqlite3.db_close db))
    with
    | Sqlite3.Error reason ->
      let msg = Printf.sprintf "opening %s: %s" filename reason in
      Sqlite_error (Sqlite3.Rc.ERROR, msg) |> raise

  let with_stmt db query fn =
    try
      let stmt = Sqlite3.prepare db query in
      Exn.protect
        ~f:begin fun () ->
          try
            fn stmt
          with
          | Sqlite3.Error reason ->
            Sqlite_error (Sqlite3.Rc.ERROR, reason) |> raise
        end
        (* TODO: would it be better to raise an exception here instead of ignore? *)
        ~finally:(fun () -> ignore (Sqlite3.finalize stmt))
    with
    | Sqlite3.Error reason ->
      let msg = Printf.sprintf "preparing '%s': %s" query reason in
      Sqlite_error (Sqlite3.Rc.ERROR, msg) |> raise

  let raise_error db =
    let code = Sqlite3.errcode db in
    let msg = Sqlite3.errmsg db in
    raise (Sqlite_error (code, msg))

  let bind_all db stmt params =
    List.iter params ~f:begin function
      | (name, value) ->
        let idx = Sqlite3.bind_parameter_index stmt name in
        match[@warning "-4"] Sqlite3.bind stmt idx value with
        | Sqlite3.Rc.OK -> ()
        | _ -> raise_error db
    end

  let each_row db stmt fn =
    let rec advance rows =
      match[@warning "-4"] Sqlite3.step stmt with
      | Sqlite3.Rc.ROW ->
        let data = Sqlite3.row_data stmt in
        let headers = Sqlite3.row_names stmt in
        let row = Array.zip_exn headers data in
        advance (Array.append rows [|fn row|])
      | Sqlite3.Rc.DONE ->
        rows
      | _ ->
        raise_error db in
    advance [||]

  let exec_fetch (type t) (module M: From_sql.Sig with type t = t) ?(params=[]) db query =
    with_stmt db query begin fun stmt ->
      if not (List.is_empty params) then bind_all db stmt params;
      each_row db stmt M.of_sql_row
    end

  let exec_only ?(params=[]) db query =
    let module Ignore = (struct
      type t = unit
      let of_sql_row = ignore
    end) in
    ignore (exec_fetch (module Ignore) ~params db query)
end
